Program Aniwemva_steepds_illu

  use sep

  ! Aniso modules
  use ani_type
  use ani_parms_cg
  use ani_exec_time
  use ani_taper
  use ani_steepds_illu_mod
  use ani_dso2

  ! Lloyd referene model building
  use lloyd_model

  implicit none

  real						        :: start,finish,start1,finish1
  real, allocatable					:: refr(:,:,:), taper(:)
  real, allocatable                                     :: dmodel(:), temp_model(:,:,:), nmodel(:,:,:), model(:,:,:)
  real, allocatable                                     :: Bimage(:,:,:), illumination(:,:,:)
  complex, allocatable       			     	:: source1(:)
  integer					        :: sh_position, iter
  integer                                               :: iz, ix, isecdim
  real, allocatable                                     :: l(:,:,:), u(:,:,:)
  double precision, allocatable                         :: fun(:)
  double precision                                      :: err, funt
  real                                                  :: lamda, lam_min, factor, tola
  real                                                  :: p_hi(2), p_lo(2)


  call sep_init()

  call aniparam()

  ! ---------------------------------------------------------------
  !             Initialization
  ! ---------------------------------------------------------------
  start=0.;finish=0.
  call aniexec_time(start)

  ! Backgournd model, in velocity 
  allocate(model(nx,nz,npar))
  call sreed("mdl", model, 4*nx*nz*npar)
  write(0,*) "initial model in velocity ", maxval(model), minval(model)

  ! Space for new model
  allocate(nmodel(nx,nz,npar)); nmodel = model;

  ! Space for model updates
  allocate(dmodel(nx*nz*npar)); dmodel=0.

  ! Space for convert 1D dmodel to 3D dmodel
  allocate(temp_model(nx,nz,npar))

  ! Bounds for model updates; bounds are converted to slowness
  allocate(l(nx,nz,npar)); l=0.
  allocate(u(nx,nz,npar)); u=0.
  call from_param('p_hi', p_hi)
  call from_param('p_lo', p_lo)
  u(:,:,1) = model(:,:,1) * (1-p_hi(1)); u(:,:,1) = 1./u(:,:,1); write(0,*) "upper bound for slow", maxval(u(:,:,1)) 
  l(:,:,1) = model(:,:,1) * (1+p_lo(1)); l(:,:,1) = 1./l(:,:,1); write(0,*) "lower bound for slow", minval(u(:,:,1)) 
  u(:,:,2) = p_hi(2) ; write(0,*) "upper bound for eta", maxval(u(:,:,2)) 
  l(:,:,2) = -p_lo(2)  ; write(0,*) "lower bound for eta", minval(l(:,:,2)) 

  ! Convert everything to slowness
  model(:,:,1) =  1./model(:,:,1)   ! slowness; eta
  write(0,*) "after conver to slowness", maxval(model), minval(model)


  ! Background image & background illumination
  allocate(Bimage(nx,nsecdim,nz)); Bimage=0.
  allocate(illumination(nx,nsecdim,nz)); illumination=0.

  ! Source wavelet
  allocate(source1(sou%w%n)); call sreed("sou", source1, 8*sou%w%n)

  ! Taper
  allocate(taper(nx)); call anitaper(tapertr,taper)

  ! Iteration information
  allocate(fun(niter+1)); fun=0;

  ! Steepest descent
  call from_param("factor", factor, 0.5)
  call from_param("lam_min", lam_min, 1.e-5)
  call from_param("tola", tola, 1.e-13)

  call aniexec_time(finish)
  write(0,'(8x,a,f10.2,a)')' Init.phase - Total elapsed time:', (finish - start),' (sec)'
  ! ---------------------------------------------------------------

  ! ---------------------------------------------------------------
  ! The main loop
  ! ---------------------------------------------------------------
  call steepds_init(source1, taper)  
  allocate(refr(nref_max,nz,npar))

  iter = 1
  err = 1

  do while (err > tola .and. iter <= niter)

    call fgeval_init(iter)

    write(0,*) "**************************************************"
    write(0,*) "         New iteration = ", iter
    write(0,*) "**************************************************"

    ! Get the backgound image for iter i
    if (iter==1) then
      refr = 0.; call lloydmodel(scales, nstart, model, refr)
      call feval(model_in=model, refr_in=refr, fun=fun(iter), Bimage=Bimage, illumination=illumination)
      call srite("bgimage", Bimage, 4*nx*nsecdim*nz)
      call srite("illum", illumination, 4*nx*nsecdim*nz)
      write(0,*) "Iter=", iter, "write the background image 1"
!      stop
    else 
      call sseek("bgimage", -4*nx*nsecdim*nz, 2)
      write(0,*) "Iter=", iter, "read the background image"
      call sreed("bgimage", Bimage, 4*nx*nsecdim*nz)
    end if

    ! Compute the perturbed image using dso
    call anidso(Bimage, mdso, illumination)

    ! Compute the gradient
    refr = 0.; call lloydmodel(scales, nstart, model, refr)
    call geval(model_in=model, refr_in=refr, Bimage=Bimage, grad=dmodel)

    ! Line Search  
    write(0,*) "starting line search"
    lamda = 0.5
    temp_model = reshape(dmodel,(/nx,nz,npar/))
    nmodel = model + lamda * temp_model

    ! Hard bounds for the new model 
    where(nmodel < l);
      nmodel = l;
    end where
    where(nmodel > u);
      nmodel = u;
    end where

    refr=0.; call lloydmodel(scales, nstart, nmodel, refr)
    call feval(model_in=nmodel, refr_in=refr, fun=funt, Bimage=Bimage, illumination=illumination)
    write(0,*) "1) fun new = ", sngl(funt), "fun = ", sngl(fun(iter)), "lamda=", lamda

    do while(funt > fun(iter)  .and. lamda > lam_min)
      lamda = lamda * factor
      nmodel = model + lamda * temp_model 
      where(nmodel < l);
        nmodel = l;
      end where
      where(nmodel > u);
        nmodel = u;
      end where
      refr=0.; call lloydmodel(scales, nstart, nmodel, refr)
      call feval(model_in=nmodel, refr_in=refr, fun=funt, Bimage=Bimage, illumination=illumination)
      write(0,*) "2) fun new=", sngl(funt), "fun kept=", sngl(fun(iter)), "lamda=", lamda
    end do
    ! ******************** 
    ! End of Line Search
    ! ******************** 
    call to_history("n4", iter+1, "bgimage")
    write(0,*) "Iter=", iter, "write the background image 2"
    call srite("bgimage", Bimage, 4*nx*nsecdim*nz)
    model = nmodel 
    fun(iter+1) = funt

    call to_history("n4", iter)
    write(0,*) "Iter=", iter, "write the new model"
    call srite("out", model, 4*nx*npar*nz)

    err  = sum(lamda*lamda*dmodel*dmodel) 
    tola = sum(nmodel*nmodel) * 1.e-10 * 1.e-10 
    write(0,*) 'Iteration = ',iter,' error=',err, ' tola=', tola
    write(0,*) "03) fn = ",funt
    iter = iter + 1
  end do

  call to_history("n1", niter, "fun")
  call to_history("d1",    1., "fun")
  call to_history("01",    1., "fun")
  call srite("fun", sngl(fun), 4*niter)

  01  format(a9,i7,a6,i9)
  02  format(a9,i7,2(a6,i9))
  03  format(a9,i7)
  04  format(a12,i4,a7,i11)

  deallocate(model, nmodel, dmodel, temp_model)
  deallocate(source1, taper, refr)
  deallocate(l, u, fun) !, sngl_model)

  call sep_close()
end Program Aniwemva_steepds_illu

