module ani_parms_cg

  use sep
  use ani_type

  implicit none

  ! Forward or adjoint, dottest or not
  logical,        public, save				:: adj,dot

  ! Define the parameter grid given by input data
  type (dat_grid), public, save	        		:: rec, sou
  type (img_grid), public, save	        		:: img
  type (modl) , public, save				:: mdl
  type (axis) , public, save     			:: h, w, xd, xm, z, sh, ksh, th, par

  ! Define the parameters for spline smoothing
  type (axis) , public, save     			:: axk_v, azk_v     
  type (axis) , public, save     			:: axk_e, azk_e     

  ! Define the useful parameters in the program
  type (axis),    public, save  			:: kxm
  integer	 ,public, save				:: tapertr,padtraces
  integer	 ,public, save				:: node,nsecdim,mdso,niter, iter_v
  integer	 ,public, save				:: nx, nz, nt, nsh, npar_in, npar_inv, nref_max
  real   	 ,public, save				:: dx, ox, dz, oz, dt, ot, dsh, osh
  real           ,public, save				:: pi, osecdim, dsecdim
  real           ,public, save				:: illu_min, illu_max 
  real,    allocatable, public, save                    :: scales(:), g_scales(:)
  integer, allocatable, public, save                    :: nstart(:)

  contains

  subroutine aniparam()

    integer						:: iw, i

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Read in parameters
    call from_param("dot",dot,.false.)
    call from_param("adj",adj,.true.)

    call from_param("node",node,9999)
    call from_param("traces_in_taper",tapertr,0)
    call from_param("padtraces",padtraces,tapertr)

    call from_param("niter", niter, 5)
    call from_param("iter_v", iter_v, 10)

    call from_param("illu_min",illu_min,  1.)
    call from_param("illu_max",illu_max, 50.)

    ! initialize file headers
    call auxin("rec")
    call auxin("sou")
    call auxin("mdl")
    call auxin("mask")

    ! Read the receiver parameter. xd is the shot position
    call from_aux("rec","n1",rec%h%n);	call from_aux("rec","d1",rec%h%d);call from_aux("rec","o1",rec%h%o)
    call from_aux("rec","n2",rec%w%n);	call from_aux("rec","d2",rec%w%d);call from_aux("rec","o2",rec%w%o)
    call from_aux("rec","n3",rec%xd%n);call from_aux("rec","d3",rec%xd%d);call from_aux("rec","o3",rec%xd%o)  

    ! Read the source parameter. 
    call from_aux("sou","n1",sou%w%n);	call from_aux("sou","d1",sou%w%d);	call from_aux("sou","o1",sou%w%o)

    ! Read the model parameters.
    call from_aux("mdl","n1",mdl%xm%n);	call from_aux("mdl","d1",mdl%xm%d);	call from_aux("mdl","o1",mdl%xm%o)
    call from_aux("mdl","n2",mdl%z%n );	call from_aux("mdl","d2",mdl%z%d );	call from_aux("mdl","o2",mdl%z%o )
    call from_aux("mdl","n3",mdl%par%n);	

    ! Read the parameters for computing reference model
    npar_in = mdl%par%n   ! # of par in the model

    if (mdl%par%n==3) then
      npar_inv = mdl%par%n-1;    ! # of par to invert
      write(0,*) "parameter # =", mdl%par%n
    else
      npar_inv = mdl%par%n
      write(0,*) "parameter # =", mdl%par%n
    end if

    call from_param("nref_max", nref_max, 3)
    allocate(scales(npar_in), nstart(npar_in),g_scales(npar_inv))
    call from_param("scales", scales)
    write(0,*) scales
    call from_param("g_scales", g_scales)
    call from_param("nstart", nstart)

    ! Read the parameters for spline smoothing
    call from_param("axkn_v", axk_v%n, 10); call from_param("axkd_v", axk_v%d, 50.); call from_param("axko_v", axk_v%o, 0.)
    call from_param("azkn_v", azk_v%n, 20); call from_param("azkd_v", azk_v%d, 8. ); call from_param("azko_v", azk_v%o, 0.)
    call from_param("axkn_e", axk_e%n, 10); call from_param("axkd_e", axk_e%d, 50.); call from_param("axko_e", axk_e%o, 0.)
    call from_param("azkn_e", azk_e%n, 20); call from_param("azkd_e", azk_e%d, 8. ); call from_param("azko_e", azk_e%o, 0.)

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Check the coherence of the source and receiver
    if (sou%w%n .ne. rec%w%n) then
      call seperr("# source freq. .ne. # receiver freq.")
    end if
    if (sou%w%d .ne. rec%w%d) then
      call seperr("source freq.interval .ne. receiver freq.interval")
    end if
    if (sou%w%o .ne. rec%w%o) then
      call seperr("source freq.origin .ne. receiver freq.origin")
    end if

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Compute the useful parameters
    pi=4*atan(1.)
    nsh=rec%xd%n; dsh=rec%xd%d; osh=rec%xd%o

    ! ----------------------------------------------------------------------------------------------------------------------
    ! The final 2-D grid that we are working on
    nz = mdl%z%n; dz = mdl%z%d; oz = mdl%z%o
    nx = rec%h%n+2*padtraces+2*tapertr  
    dx = rec%h%d
    ox = rec%h%o-(padtraces+tapertr)*dx 
    img%xm%n=nx; img%xm%o=ox; img%xm%d=dx
    img%z%n =nz; img%z%o =oz; img%z%d =dz  

    if (nx .ne. mdl%xm%n ) then
      write(0,*) "nx=", nx, "mdl%xm%n=", mdl%xm%n
      call seperr("# of working nx .ne. # of model nx")
    end if
    if (ox .ne. mdl%xm%o ) then
      write(0,*) ox, mdl%xm%o
      call seperr("Min.x-image space .ne. Min.x-model")
    end if
    if (dx .ne. mdl%xm%d ) call seperr("Sampling of working grid .ne. Sampling of model gird")    

    !--------------------------------------------------
    ! Subsurface offset vector for the image
    !
    call from_param("nh",nsecdim,1)
    call from_param("dh",dsecdim,img%xm%d)
    call from_param("oh",osecdim,-int(nsecdim/2)*dsecdim)
    call from_param("mdso", mdso, 1)
    write(0,*) "s_offset", "nh=", nsecdim, "dh=", dsecdim, "oh=", osecdim

    !--------------------------------------------------
    ! Spatial Frequency vector
    !
    kxm%n = rec%h%n+2*(tapertr+padtraces)
    kxm%d = 1./(kxm%n*rec%h%d)
    kxm%o = 1./(2*rec%h%d)-kxm%d
    allocate(kxm%all(kxm%n)); kxm%all=0.
    do i=1,kxm%n/2+1
      kxm%all(i) = (i-1) * kxm%d
    end do
    do i=kxm%n/2+2, kxm%n
      kxm%all(i) = (i-kxm%n-1) * kxm%d
    end do
    kxm%all = kxm%all * pi*2

    !--------------------------------------------------
    ! Frequency vector
    !
    rec%w%o   = 2*pi * rec%w%o 
    rec%w%d   = 2*pi * rec%w%d  
    allocate(rec%w%all(rec%w%n))
    do iw = 1,rec%w%n
      rec%w%all(iw) = rec%w%o + (iw - 1) * rec%w%d
    end do

    !--------------------------------------------------
    ! Time axis
    ! FFt from frequency
    nt = rec%w%n * 2
    dt = 1./(rec%w%n*nt) * pi*2
    ot = 0.

    ! -------------------------------------------------
    ! works for symmetric offset

    call auxout("grad"); call auxout("grad_ns"); call auxinout("bgimage"); call auxout("illum")

    call to_history("n1",nx,"grad") ; call to_history("d1",dx,"grad") ; call to_history("o1",ox,"grad")
    call to_history("n2",nz,"grad") ; call to_history("d2",dz,"grad") ; call to_history("o2",oz,"grad")
    call to_history("n3",npar_inv,"grad") ; call to_history("d3",1,"grad" ) ; call to_history("o3",0 ,"grad")
    call to_history("label2", "z(m)","grad"); call to_history("label1", "x(m)","grad"); 
    call to_history("label3", "par","grad"); call to_history("label4", "iteration");
    call to_history("esize",4,"grad")

    call to_history("n1",nx,"grad_ns") ; call to_history("d1",dx,"grad_ns") ; call to_history("o1",ox,"grad_ns")
    call to_history("n2",nz,"grad_ns") ; call to_history("d2",dz,"grad_ns") ; call to_history("o2",oz,"grad_ns")
    call to_history("n3",npar_inv,"grad_ns") ; call to_history("d3",1,"grad_ns" ) ; call to_history("o3",0 ,"grad_ns")
    call to_history("label2", "z(m)","grad_ns"); call to_history("label1", "x(m)","grad_ns"); 
    call to_history("label3", "par","grad_ns"); call to_history("label4", "iteration");
    call to_history("esize",4,"grad_ns")

    call to_history("n1",nx,"illum") ; call to_history("d1",dx,"illum") ; call to_history("o1",ox,"illum")
    call to_history("n2",nz,"illum") ; call to_history("d2",dz,"illum") ; call to_history("o2",oz,"illum")
    call to_history("label2", "z(m)","illum"); call to_history("label1", "x(m)","illum"); 
    call to_history("esize",4,"illum")

    call to_history("n1",nx,"bgimage") ;         call to_history("d1",dx,"bgimage") ;         call to_history("o1",ox,"bgimage")
    call to_history("n2",nsecdim,"bgimage") ;    call to_history("d2",dsecdim,"bgimage" ) ;   call to_history("o2",osecdim,"bgimage")
    call to_history("n3",nz,"bgimage") ;         call to_history("d3",dz,"bgimage") ;         call to_history("o3",oz,"bgimage")
    call to_history("d4",1 ,"bgimage") ;         call to_history("o4",0 ,"bgimage")
    call to_history("label3", "z(m)","bgimage"); call to_history("label1", "x(m)","bgimage");
    call to_history("label2", "soffset","bgimage"); call to_history("label4", "iteration");
    call to_history("esize",4,"bgimage")

    call to_history("n1",nx) ;   call to_history("d1",dx) ; call to_history("o1",ox)
    call to_history("n2",nz) ;   call to_history("d2",dz) ; call to_history("o2",oz)
    call to_history("n3",npar_in) ; call to_history("d3",1 ) ; call to_history("o3",0 )
    call to_history("label2", "z(m)"); call to_history("label1", "x(m)");
    call to_history("label3", "par"); call to_history("label4", "iteration"); 
    call to_history("esize",4); call to_history("n4", niter)
  end subroutine aniparam
end module ani_parms_cg
