module ani_wemvacg_illu

  use sep
  use ani_type
  use ani_fftw
  use ani_parms_cg
  use ani_image
  use eta_pspi
  use eta_phsscr
  use adj_mod
  use ani_exec_time

  implicit none

  integer, private                                       :: idip, nodes
  complex, allocatable, private                          :: source1(:)
  real, allocatable, private                             :: taper(:), model(:,:,:), refr(:,:,:)

  contains

  !======================================================================
  subroutine aniwei_init()
    integer, external			:: omp_get_thread_num,omp_get_num_threads
    if (node.eq.9999) then
      !$OMP PARALLEL
      nodes = omp_get_num_threads()
      !$OMP END PARALLEL
    else
      nodes=node
    end if
    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"

  end subroutine aniwei_init

  !======================================================================
  subroutine aniwei(taper,recv,source,model_in,refr_in,image,illum)

    complex						:: recv(:,:),source(:,:)
    integer						:: iwa,iz,inode,i,ixw,isecdim
    real						:: model_in(:,:,:), refr_in(:,:,:)
    real						:: image(:,:,:), taper(:)
    real, allocatable					:: GGG(:,:,:,:), III(:,:,:,:)
    real, allocatable					:: model(:,:,:), refr(:,:,:)
    real, optional                                      :: illum(:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: tflds(:,:),tfldr(:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads

    allocate(ws(nx,nodes), wr(nx,nodes)); wr=0.; ws=0.
    allocate(tflds(nx,nodes), tfldr(nx,nodes)); tflds=0.;tfldr=0.
    allocate(GGG(nx,nsecdim,nz,nodes)); GGG=0.
    allocate(model(nx, nz, npar), refr(nref_max, nz, npar)); 
    if (present(illum)) then
      allocate(III(nx,nsecdim,nz,nodes)); III=0.
    end if
    model = model_in; refr  = refr_in

    ! Setup the fftw plans 
    call ani_fftws(tflds, nodes)
    call ani_fftwr(tfldr, nodes)

    !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
    do iwa = 1, rec%w%n
      iz = 1; inode=omp_get_thread_num()+1
      ws(:,inode)=0.;wr(:,inode)=0.
      ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)
      wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
      if (present(illum)) then
        call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode), sillu=III(:,:,iz,inode))   
      else
        call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode))   
      end if
      do iz = 2, nz
        call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
        call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
        ws(:,inode)=ws(:,inode)*taper
        wr(:,inode)=wr(:,inode)*taper
        if (present(illum)) then
          call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode), sillu=III(:,:,iz,inode))
        else
          call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GGG(:,:,iz,inode))
        end if
      end do
    end do
    !$OMP END PARALLEL DO

    do inode=1,nodes
      !$OMP parallel do private(ixw,iz,isecdim)
      do iz=1,nz
        do isecdim=1,nsecdim
          do ixw=1,kxm%n
            if (present(illum)) then
              illum(ixw,isecdim,iz) = illum(ixw,isecdim,iz) + III(ixw,isecdim,iz,inode)
            end if
            image(ixw,isecdim,iz)=image(ixw,isecdim,iz)+GGG(ixw,isecdim,iz,inode)
          end do
        end do
      end do
      !$OMP END PARALLEL DO
    end do
    6   format(A13,i3,A8,i3,A14,i3,A12,i3)
    deallocate(ws, wr, GGG, model, refr, tflds, tfldr )
    if(allocated(III)) deallocate(III)
  end subroutine aniwei




  !======================================================================
  subroutine aniwemva_init(src_in, taper_in, model_in, refr_in)

    integer                             :: idip_in, nodes
    integer, external			:: omp_get_thread_num,omp_get_num_threads
    complex                             :: src_in(:)
    real                                :: taper_in(:), model_in(:,:,:), refr_in(:,:,:)
    if (node.eq.9999) then
      !$OMP PARALLEL
      nodes = omp_get_num_threads()
      !$OMP END PARALLEL
    else
      nodes=node
    end if

    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)
    if (allocated(model))  deallocate(model)
    if (allocated(refr))   deallocate(refr)

    allocate(source1(sou%w%n), taper(nx))
    allocate(model(nx,nz,npar), refr(nref_max,nz,npar))

    source1 = src_in
    taper  = taper_in
    model  = model_in
    refr   = refr_in

  end subroutine aniwemva_init
  !======================================================================

  !======================================================================
  function aniwemva_lop(adj, add, modl, data) result(stat)

    integer                :: stat
    logical, intent(in)    :: adj, add
    real, dimension(:)     :: modl, data

    stat = 1
    call adjnull(adj, add, modl, data)
    call aniwemva_lop2(adj, add, modl, data)
    stat = 0

  end function aniwemva_lop
  !======================================================================

  !======================================================================
  subroutine aniwemva_lop2(adj, add, deltaM, deltaI)
    logical						:: adj, add
    integer						:: iwa,iz,inode,isecdim,ixw,ish
    real        					:: deltaM(nx,nz,npar)
    real	        				:: deltaI(nx,nsecdim,nz)
    real, allocatable					:: SSS(:,:,:), EEE(:,:,:)
    real, allocatable					:: GGG(:,:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: tflds(:,:),tfldr(:,:)
    complex, allocatable                                :: source(:,:), recv(:,:)
    complex, allocatable				:: ddws(:,:),ddwr(:,:)
    complex, allocatable				:: dws(:,:,:),dwr(:,:,:)
    complex, allocatable				:: bws(:,:,:),bwr(:,:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads
    real                                                :: start, finish, start1, finish1
    integer                                             :: sh_position

    allocate(bws(nx,nz,nodes), bwr(nx,nz,nodes)); bwr=0; bws=0
    allocate(ws(nx,nodes), wr(nx,nodes)); ws=0; wr=0
    allocate(tflds(nx,nodes), tfldr(nx,nodes)); tflds=0.;tfldr=0.
    allocate(ddws(nx,nodes), ddwr(nx,nodes)); ddws=0; ddwr=0
    if(.not. allocated(source)) allocate(source(rec%h%n, sou%w%n))
    if(.not. allocated(recv))   allocate(recv  (rec%h%n, rec%w%n))

    if (adj) then
      deltaM=0.
    else
      deltaI=0.
    end if

    ! Setup the fftw plans 
    call ani_fftws(tflds, nodes)
    call ani_fftwr(tfldr, nodes)

    call sseek("rec",0,0)

    ! ----------------------------------------
    ! Shot loop
    do ish = 1, nsh
      start=0; finish=0; source = 0.; recv=0
      call aniexec_time(start)
      sh_position = ((ish-1)*dsh + osh - rec%h%o) / dx + 1
      !      write(0,*) "shot position=", sh_position, (ish-1)*dsh + osh
      source(sh_position,:) = conjg(source1);
      call sreed("rec", recv, 8*rec%h%n*rec%w%n)
      recv = conjg(recv);

      ! ----------------------------------------
      ! Adjoint tomographic operator
      if (adj) then
        allocate(SSS(nx,nz,nodes),EEE(nx,nz,nodes))
        SSS=0.                                 ! Temporary location for slowness perturbation
        EEE=0.                                 ! Temporary location for \eta perturbation 
        allocate(dwr(1,kxm%n,nodes));     dwr=0.
        allocate(dws(1,kxm%n,nodes));     dws=0.

        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)       
        do iwa = 1, rec%w%n
          ! Initializing 
          inode=omp_get_thread_num()+1; 
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)     ! Source and recv are not padded
          wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
          dws(1,:,inode) = 0.
          dwr(1,:,inode) = 0.
          bws(:,:,inode) = 0.
          bwr(:,:,inode) = 0.
          bws(:,1,inode) = ws(:,inode)
          bwr(:,1,inode) = wr(:,inode)

          ! Background wavefields, bws & bwr
          do iz = 2, nz
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=nz, 2, -1
            ! Adjoint imaging condition to get the wavefield perturbation
            call aniimage(.true., iwa, dws(1,:,inode), bwr(:,iz,inode), deltaI(:,:,iz), ssou=.true. )
            call aniimage(.true., iwa, bws(:,iz,inode), dwr(1,:,inode), deltaI(:,:,iz), ssou=.false.)

            ! Upward propagate wavefield perturbations
            call etapspi(iwa, inode,  1, dws(1,:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
            dws(1,:,inode)=dws(1,:,inode)*taper
            call etapspi(iwa, inode, -1, dwr(1,:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Phase-screen operator & adjoint scattering operator to get the model perturbation
            ddws(:,inode)=0
            call aniphsscr_slow(iwa,-1,ddws(:,inode),bws(:,iz-1,inode),model(:,iz-1,1),model(:,iz-1,2))
            SSS(:,iz-1,inode)=SSS(:,iz-1,inode)-conjg(ddws(:,inode))*(dws(1,:,inode))/rec%w%n    

            ddws(:,inode)=0
            call aniphsscr_eta (iwa,-1,ddws(:,inode),bws(:,iz-1,inode),model(:,iz-1,1))
            EEE(:,iz-1,inode)=EEE(:,iz-1,inode)-conjg(ddws(:,inode))*(dws(1,:,inode))/rec%w%n    

            ddwr(:,inode)=0
            call aniphsscr_slow(iwa, 1,ddwr(:,inode),bwr(:,iz-1,inode),model(:,iz-1,1),model(:,iz-1,2))
            SSS(:,iz-1,inode)=SSS(:,iz-1,inode)-conjg(ddwr(:,inode))*(dwr(1,:,inode))/rec%w%n

            ddwr(:,inode)=0
            call aniphsscr_eta (iwa, 1,ddwr(:,inode),bwr(:,iz-1,inode),model(:,iz-1,1))
            EEE(:,iz-1,inode)=EEE(:,iz-1,inode)-conjg(ddwr(:,inode))*(dwr(1,:,inode))/rec%w%n  
          end do

          !              if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
          !                   " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz)
          do iz=1,nz
            do ixw=1,kxm%n
              deltaM(ixw,iz,1)=deltaM(ixw,iz,1)+SSS(ixw,iz,inode)
              deltaM(ixw,iz,2)=deltaM(ixw,iz,2)+EEE(ixw,iz,inode)
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,EEE,SSS)
        ! ----------------------------------------
        ! Forward tomographic operator
      else 
        allocate(GGG(nx,nsecdim,nz,nodes))
        GGG=0.
        allocate(dwr(1,kxm%n,nodes));               dwr=0.
        allocate(dws(1,kxm%n,nodes));               dws=0.
        !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)

        do iwa = 1,rec%w%n
          inode=omp_get_thread_num()+1;
          ws(:,inode)=0.;wr(:,inode)=0.
          ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)
          wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
          dws(1,:,inode)=0.
          dwr(1,:,inode)=0.
          bws(:,:,inode)=0.
          bwr(:,:,inode)=0.
          bws(:,1,inode)=ws(:,inode)
          bwr(:,1,inode)=wr(:,inode)

          do iz = 2, nz
            call etapspi(iwa, inode, -1, ws(:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "sou")
            call etapspi(iwa, inode,  1, wr(:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:), "rec")
            ws(:,inode)=ws(:,inode)*taper
            wr(:,inode)=wr(:,inode)*taper
            bws(:,iz,inode)=ws(:,inode)
            bwr(:,iz,inode)=wr(:,inode)
          end do

          do iz=2,nz
            ! Phase-screen
            ddws(:,inode) = 0
            call aniphsscr_slow(iwa, 1,ddws(:,inode),bws(:,iz-1,inode),model(:,iz-1,1),model(:,iz-1,2))
            dws(1,:,inode)=dws(1,:,inode)+(ddws(:,inode))*deltaM(:,iz-1,1)/rec%w%n   

            ddws(:,inode)=0
            call aniphsscr_eta (iwa, 1,ddws(:,inode),bws(:,iz-1,inode),model(:,iz-1,1))
            dws(1,:,inode)=dws(1,:,inode)+(ddws(:,inode))*deltaM(:,iz-1,2)/rec%w%n    

            ! Scattering
            call etapspi(iwa, inode,-1, dws(1,:,inode), tflds(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"sou")
            dws(1,:,inode)=dws(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, dws(1,:,inode), bwr(:,iz,inode), GGG(iz,:,:,inode))

            ! Phase-screen
            ddwr(:,inode)=0
            call aniphsscr_slow(iwa,-1,ddwr(:,inode),bwr(:,iz-1,inode),model(:,iz-1,1),model(:,iz-1,2))
            dwr(1,:,inode)=dwr(1,:,inode)+(ddwr(:,inode))*deltaM(:,iz-1,1)/rec%w%n    

            ddwr(:,inode)=0
            call aniphsscr_eta (iwa,-1,ddwr(:,inode),bwr(:,iz-1,inode),model(:,iz-1,1))
            dwr(1,:,inode)=dwr(1,:,inode)+(ddwr(:,inode))*deltaM(:,iz-1,2)/rec%w%n 

            ! Scattering
            call etapspi(iwa, inode, 1, dwr(1,:,inode), tfldr(:,inode), model(:,iz-1,:), refr(:,iz-1,:),"rec")
            dwr(1,:,inode)=dwr(1,:,inode)*taper

            ! Imaging condition
            call aniimage(.false., iwa, bws(:,iz,inode), dwr(1,:,inode), GGG(:,:,iz,inode))
          end do

          if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
          " FROM NODE: ",inode
        end do
        !$OMP END PARALLEL DO

        do inode=1,nodes
          !$OMP parallel do private(ixw,iz,isecdim)
          do iz=1,nz
            do isecdim=1,nsecdim
              do ixw=1,kxm%n
                deltaI(ixw,isecdim,iz)=deltaI(ixw,isecdim,iz)+GGG(ixw,isecdim,iz,inode)
              end do
            end do
          end do
          !$OMP END PARALLEL DO
        end do
        deallocate(dwr,dws,GGG)

      end if
    end do
    deallocate(bws,bwr,ws,wr,ddws,ddwr,tflds,tfldr)
    deallocate(source, recv)
    6   format(A13,i3,A8,i3,A14,i3,A12,i3)

  end subroutine aniwemva_lop2
  !======================================================================

  !======================================================================
  subroutine aniwemva_close()

    if (allocated(source1)) deallocate(source1)
    if (allocated(taper))  deallocate(taper)
    if (allocated(model))  deallocate(model)
    if (allocated(refr))   deallocate(refr)

  end subroutine aniwemva_close
  !======================================================================

end module ani_wemvacg_illu
